import os

from twisted.internet.defer import inlineCallbacks
from yaml import dump

from ensemble.control import main
from ensemble.errors import FileNotFound
from ensemble.environment.environment import Environment
from ensemble.formula.repository import LocalFormulaRepository

from ensemble.unit.workflow import UnitWorkflowState

from .common import MachineControlToolTest


class FormulaUpgradeTestBase(object):

    def add_formula(self, metadata, repository_dir=None):
        if repository_dir is None:
            repository_dir = self.makeDir()
        formula_dir = self.makeDir(dirname=repository_dir)
        fh = open(os.path.join(formula_dir, "metadata.yaml"), "w")
        fh.write(dump(metadata))
        fh.close()
        return LocalFormulaRepository(repository_dir)

    def increment_formula(self, formula):
        metadata = formula.metadata.get_serialization_data()
        metadata["name"] = "mysql"
        metadata["revision"] = 2
        repository = self.add_formula(metadata)
        return repository


class ControlFormulaUpgradeTest(
    MachineControlToolTest, FormulaUpgradeTestBase):

    @inlineCallbacks
    def setUp(self):
        yield super(ControlFormulaUpgradeTest, self).setUp()
        config = {
            "ensemble": "environments",
            "environments": {"firstenv": {"type": "dummy"}}}

        self.write_config(dump(config))
        self.config.load()
        self.service_state1 = yield self.add_service_from_formula("mysql")
        self.service_unit1 = yield self.service_state1.add_unit_state()

        self.unit1_workflow = UnitWorkflowState(
            self.client, self.service_unit1, None, self.makeDir())
        yield self.unit1_workflow.set_state("started")

        self.environment = self.config.get_default()
        self.provider = self.environment.get_machine_provider()

        self.output = self.capture_logging()
        self.stderr = self.capture_stream("stderr")

    @inlineCallbacks
    def test_formula_upgrade(self):
        """
        'ensemble formula-upgrade <service_name>' will schedule
        a formula for upgrade.
        """
        repository = self.increment_formula(self.formula)

        mock_environment = self.mocker.patch(Environment)
        mock_environment.get_machine_provider()
        self.mocker.result(self.provider)

        finished = self.setup_cli_reactor()
        self.setup_exit(0)
        self.mocker.replay()
        main(["upgrade-formula", "--repository", repository.path, "mysql"])
        yield finished

        # Verify the service has a new formula reference
        formula_id = yield self.service_state1.get_formula_id()
        self.assertEqual(formula_id, "local:mysql-2")

        # Verify the provider storage has been updated
        formula = repository.find("mysql")
        storage = self.provider.get_file_storage()
        try:
            yield storage.get(
                "formulas/local:mysql-2:%s" % formula.get_sha256())
        except FileNotFound:
            self.fail("New formula not uploaded")

        # Verify the upgrade flag on the service units.
        upgrade_flag = yield self.service_unit1.get_upgrade_flag()
        self.assertTrue(upgrade_flag)

    @inlineCallbacks
    def test_upgrade_formula_with_unupgradeable_units(self):
        """If there are units that won't be upgraded, they will be reported,
        other units will be upgraded.
        """
        repository = self.increment_formula(self.formula)
        service_unit2 = yield self.service_state1.add_unit_state()

        finished = self.setup_cli_reactor()
        self.setup_exit(0)
        self.mocker.replay()

        main(["upgrade-formula", "--repository", repository.path, "mysql"])
        yield finished

        # Verify report of unupgradeable units
        self.assertIn(
            ("Unit 'mysql/1' is not in a running state "
            "(state: 'uninitialized'), won't upgrade"),
            self.output.getvalue())

        # Verify flags only set on upgradeable unit.
        value = (yield service_unit2.get_upgrade_flag())
        self.assertFalse(value)
        value = (yield self.service_unit1.get_upgrade_flag())
        self.assertTrue(value)

    @inlineCallbacks
    def test_upgrade_formula_unknown_service(self):
        finished = self.setup_cli_reactor()
        self.setup_exit(0)
        self.mocker.replay()
        main(["upgrade-formula", "--repository", self.makeDir(), "volcano"])
        yield finished
        self.assertIn(
            "Service 'volcano' was not found", self.stderr.getvalue())

    @inlineCallbacks
    def test_upgrade_formula_unknown_formula(self):
        """If a formula is not found in the repository, an error is given.
        """
        finished = self.setup_cli_reactor()
        self.setup_exit(0)
        self.mocker.replay()
        repository_dir = self.makeDir()
        main(["upgrade-formula", "--repository", repository_dir, "mysql"])
        yield finished
        self.assertIn(
            "Formula 'mysql' not found in repository", self.output.getvalue())

    @inlineCallbacks
    def test_upgrade_formula_dryrun(self):
        """If a formula is not found in the repository, an error is given.
        """
        finished = self.setup_cli_reactor()
        self.setup_exit(0)
        self.mocker.replay()
        repository_dir = self.makeDir()
        main(["upgrade-formula", "--repository", repository_dir, "mysql"])
        yield finished
        self.assertIn(
            "Formula 'mysql' not found in repository", self.output.getvalue())

    @inlineCallbacks
    def test_upgrade_formula_dryrun_reports_unupgradeable_units(self):
        """If there are units that won't be upgraded, dry-run will report them.
        """
        repository = self.increment_formula(self.formula)
        service_unit2 = yield self.service_state1.add_unit_state()

        finished = self.setup_cli_reactor()
        self.setup_exit(0)
        self.mocker.replay()

        main(["upgrade-formula", "-n",
              "--repository", repository.path, "mysql"])
        yield finished

        # Verify dry run
        self.assertIn(
            "Service would be upgraded from formula", self.output.getvalue())
        # Verify report of unupgradeable units
        self.assertIn(
            ("Unit 'mysql/1' is not in a running state "
            "(state: 'uninitialized'), won't upgrade"),
            self.output.getvalue())

        # Verify no flags have been set.
        value = (yield service_unit2.get_upgrade_flag())
        self.assertFalse(value)
        value = (yield self.service_unit1.get_upgrade_flag())
        self.assertFalse(value)

    @inlineCallbacks
    def test_upgrade_formula_dryrun_latest(self):
        """Dry run mode outputs if the latest formula is deployed.
        """
        finished = self.setup_cli_reactor()
        self.setup_exit(0)
        self.mocker.replay()

        metadata = self.formula.metadata.get_serialization_data()
        metadata["name"] = "mysql"
        repository = self.add_formula(metadata)
        main(["upgrade-formula", "--dry-run",
                  "--repository", repository.path, "mysql"])
        yield finished
        self.assertIn(
            "Service already running latest formula",
            self.output.getvalue())

        upgrade_flag = yield self.service_unit1.get_upgrade_flag()
        self.assertFalse(upgrade_flag)

    @inlineCallbacks
    def test_upgrade_formula_service_using_latest(self):
        """If the service is already running the latest formula no work is done

        An appropriate output noting this condition is given.
        """
        finished = self.setup_cli_reactor()
        self.setup_exit(0)
        self.mocker.replay()

        metadata = self.formula.metadata.get_serialization_data()
        metadata["name"] = "mysql"
        repository = self.add_formula(metadata)
        main(["upgrade-formula", "--repository", repository.path, "mysql"])

        yield finished
        self.assertIn(
            "Formula 'namespace:mysql-1' is the latest revision known",
            self.output.getvalue())

        upgrade_flag = yield self.service_unit1.get_upgrade_flag()
        self.assertFalse(upgrade_flag)
