import os
import stat
import hashlib
import inspect
import zipfile

from ensemble.lib.testing import TestCase
from ensemble.lib.filehash import compute_file_hash
from ensemble.formula.metadata import MetaData
from ensemble.formula.bundle import FormulaBundle
from ensemble.errors import FormulaError
from ensemble.formula.directory import FormulaDirectory
from ensemble.formula.provider import get_formula_from_path

from ensemble.formula import tests

repository_directory = os.path.join(
    os.path.dirname(inspect.getabsfile(tests)), "repository")

sample_directory = os.path.join(repository_directory, "dummy")


class BundleTest(TestCase):

    def setUp(self):
        directory = FormulaDirectory(sample_directory)

        # add sample directory
        self.filename = self.makeFile(suffix=".frml")
        directory.make_archive(self.filename)

    def test_initialization(self):
        bundle = FormulaBundle(self.filename)
        self.assertEquals(bundle.path, self.filename)

    def test_error_not_zip(self):
        filename = self.makeFile("@#$@$")
        try:
            FormulaBundle(filename)
        except FormulaError, exc:
            self.assertIn(filename, str(exc))
            self.assertIn("zip", str(exc).lower())
            return

        self.fail("Expected formula error.")

    def test_error_zip_but_doesnt_have_metadata_file(self):
        filename = self.makeFile()
        zf = zipfile.ZipFile(filename, 'w')
        zf.writestr("README.txt", "This is not a valid formula.")
        zf.close()

        try:
            FormulaBundle(filename)
        except FormulaError, exc:
            self.assertIn(filename, str(exc))
            self.assertIn("metadata.yaml", str(exc).lower())
            return

        self.fail("Expected formula error.")

    def test_bundled_config(self):
        """Make sure that config is accessible from a bundle."""
        from ensemble.formula.tests.test_config import sample_yaml_data
        bundle = FormulaBundle(self.filename)
        self.assertEquals(bundle.config.get_serialization_data(),
                          sample_yaml_data)

    def test_info(self):
        bundle = FormulaBundle(self.filename)
        self.assertTrue(bundle.metadata is not None)
        self.assertTrue(isinstance(bundle.metadata, MetaData))
        self.assertEquals(bundle.metadata.name, "dummy")

    def test_as_bundle(self):
        bundle = FormulaBundle(self.filename)
        self.assertEquals(bundle.as_bundle(), bundle)

    def test_executable_extraction(self):
        sample_directory = os.path.join(repository_directory, "varnish2")
        formula_directory = FormulaDirectory(sample_directory)
        bundle = formula_directory.as_bundle()
        directory = bundle.as_directory()
        hook_path = os.path.join(directory.path, "hooks", "install")
        self.assertEqual(stat.S_IMODE(os.stat(hook_path)[0]), 493)

    def get_formula_sha256(self):
        return compute_file_hash(hashlib.sha256, self.filename)

    def test_compute_sha256(self):
        sha256 = self.get_formula_sha256()
        bundle = FormulaBundle(self.filename)
        self.assertEquals(bundle.compute_sha256(), sha256)

    def test_formula_base_inheritance(self):
        """
        get_sha256() should be implemented in the base class,
        and should use compute_sha256 to calculate the digest.
        """
        sha256 = self.get_formula_sha256()
        bundle = FormulaBundle(self.filename)
        self.assertEquals(bundle.get_sha256(), sha256)

    def test_file_handle_as_path(self):
        sha256 = self.get_formula_sha256()
        fh = open(self.filename)
        bundle = FormulaBundle(fh)
        self.assertEquals(bundle.get_sha256(), sha256)

    def test_extract_to(self):
        filename = self.makeFile()
        formula = get_formula_from_path(self.filename)
        f2 = formula.extract_to(filename)

        # f2 should be a formula directory
        self.assertInstance(f2, FormulaDirectory)
        self.assertInstance(f2.get_sha256(), basestring)
        self.assertEqual(f2.path, filename)

    def test_as_directory(self):
        filename = self.makeFile()
        formula = get_formula_from_path(self.filename)
        f2 = formula.as_directory()

        # f2 should be a formula directory
        self.assertInstance(f2, FormulaDirectory)
        self.assertInstance(f2.get_sha256(), basestring)
        # verify that it was extracted to a new temp dirname
        self.assertNotEqual(f2.path, filename)

        fn = os.path.split(f2.path)[1]
        # verify that it used the expected prefix
        self.assertStartsWith(fn, "tmp")
