# -*- encoding: utf-8 -*-
import yaml

from ensemble.lib.testing import TestCase
from ensemble.formula.config import ConfigOptions
from ensemble.formula.errors import ServiceConfigError

sample_configuration = """
options:
  title: {default: My Title, description: A descriptive title used for the service., type: str}
  outlook: {description: No default outlook., type: str}
  username: {default: admin, description: The name of the initial account (given admin permissions)., type: regex, validator: '\w{8}'}
  skill-level: {description: A number indicating skill., type: int}
"""

sample_yaml_data = yaml.load(sample_configuration)

sample_config_defaults = {
    "title": "My Title",
    "username": "admin"
    }

class ConfigOptionsTest(TestCase):

    def setUp(self):
        self.config = ConfigOptions()

    def test_load(self):
        """Validate we can load data or get expected errors."""

        # load valid data
        filename = self.makeFile(sample_configuration)
        self.config.load(filename)
        self.assertEqual(self.config.get_serialization_data(),
                         sample_yaml_data)

        # test with dict based data
        self.config.parse(sample_yaml_data)
        self.assertEqual(self.config.get_serialization_data(),
                         sample_yaml_data)


        # and with an unhandled type
        self.assertRaises(TypeError, self.config.load, 1.234)

    def test_load_file(self):
        sample_path = self.makeFile(sample_configuration)
        config = ConfigOptions()
        config.load(sample_path)

        self.assertEqual(config.get_serialization_data(),
                         sample_yaml_data)

        # and an expected exception
        # on an empty file
        empty_file = self.makeFile("")
        error = self.assertRaises(ServiceConfigError, config.load, empty_file)
        self.assertEqual(error.message,
                         "Missing required service options metadata: %s" % empty_file)

        # a missing filename is allowed
        config = config.load("missing_file")


    def test_defaults(self):
        self.config.parse(sample_configuration)
        defaults = self.config.get_defaults()
        self.assertEqual(defaults, sample_config_defaults)

    def test_parse(self):
        """Verify that parse checks and raises."""
        # no options dict
        self.assertRaises(ServiceConfigError, self.config.parse, {"foo": "bar"})

        # and with bad data expected exceptions
        self.assertRaises(ServiceConfigError,
                          self.config.parse, "foo: [1, 2")


    def test_validate(self):
        sample_input = {
            "title": "Helpful Title",
            "outlook": "Peachy",
            }

        self.config.parse(sample_configuration)
        data = self.config.validate(sample_input)

        # This should include an overridden value, a default and a new
        # value.
        self.assertEqual(data,
                         {"username": "admin",
                          "outlook": "Peachy",
                          "title": "Helpful Title"})

        # now try to set a value outside the expected
        sample_input["bad"] = "value"
        error = self.assertRaises(ServiceConfigError, self.config.validate, sample_input)
        self.assertEqual(error.message,
                         "bad is not a valid configuration option.")

        # validating with an empty instance
        # the service takes no options
        config = ConfigOptions()
        self.assertRaises(ServiceConfigError, config.validate, sample_input)

    def test_validate_regex(self):
        self.config.parse(sample_configuration)

        error = self.assertRaises(ServiceConfigError,
                                  self.config.validate, dict(username="1234"))
        self.assertIn(str(error), "Invalid value for username: 1234")


        data = self.config.validate(dict(username="12345678"))
        self.assertEqual(data, {"username": "12345678", "title": "My Title"})


    def test_validate_regex_wo_validator(self):
        """Assure that validator is required for type=regex."""
        local_configuration = sample_configuration + \
          """  fails: {description: missing required., type:  regex}"""
        error = self.assertRaises(ServiceConfigError,
                          self.config.parse,
                          local_configuration)
        self.assertIn("options.fails.validator: required value not found",
                      str(error))


    def test_validate_types(self):

        self.config.parse(sample_configuration)

        error = self.assertRaises(ServiceConfigError,
                                  self.config.validate, {"skill-level": "NaN"})
        self.assertIn(str(error), "Invalid value for skill-level: NaN")


        data = self.config.validate({"skill-level": "9001"})
        # its over 9000!
        self.assertEqual(data, {"skill-level": 9001,
                                "title": "My Title",
                                "username": "admin"})
