import os
import yaml
import zookeeper

from twisted.internet.defer import inlineCallbacks, fail
from twisted.python.failure import Failure

from txzookeeper import ZookeeperClient
from txzookeeper.tests.utils import deleteTree

from ensemble.formula.publisher import FormulaPublisher
from ensemble.formula.bundle import FormulaBundle
from ensemble.formula.directory import FormulaDirectory
from ensemble.providers.dummy import FileStorage
from ensemble.state.formula import FormulaStateManager
from ensemble.state.errors import StateChanged
from ensemble.tests.common import get_test_zookeeper_address

from ensemble.environment.tests.test_config import (
    EnvironmentsConfigTestBase, SAMPLE_ENV)
from ensemble.lib.mocker import MATCH

from .test_repository import RepositoryTestBase


class FormulaPublisherTest(RepositoryTestBase):

    @inlineCallbacks
    def setUp(self):
        super(FormulaPublisherTest, self).setUp()
        zookeeper.set_debug_level(0)

        self.formula = FormulaDirectory(self.sample_dir1)
        # formula state manager, formula id
        self.formula_key = "local:%s-%s"%(
            self.formula.metadata.name, self.formula.metadata.revision)
        # provider storage key
        self.formula_storage_key = "%s:%s" % (
            self.formula_key, self.formula.get_sha256())

        self.client = ZookeeperClient(get_test_zookeeper_address())
        self.storage = FileStorage(self.makeDir())
        self.publisher = FormulaPublisher(self.client, self.storage)

        yield self.client.connect()
        yield self.client.create("/formulas")

    def tearDown(self):
        deleteTree("/", self.client.handle)
        self.client.close()
        super(FormulaPublisherTest, self).tearDown()

    @inlineCallbacks
    def test_add_formula_and_publish(self):
        yield self.publisher.add_formula("local", self.formula)
        yield self.publisher.publish()
        children = yield self.client.get_children("/formulas")
        self.assertIn(self.formula_key, children)
        fh = yield self.storage.get("formulas/%s" % self.formula_storage_key)
        bundle = FormulaBundle(fh)
        self.assertEqual(self.formula.get_sha256(), bundle.get_sha256())

    @inlineCallbacks
    def test_published_formula_sans_unicode(self):
        yield self.publisher.add_formula("local", self.formula)
        yield self.publisher.publish()
        data, stat = yield self.client.get("/formulas/%s" % self.formula_key)
        self.assertNotIn("unicode", data)

    @inlineCallbacks
    def test_add_formula_with_concurrent(self):
        """
        Publishing a formula, that has become published concurrent, after the
        add_formula, works fine. it will write to storage regardless. The use
        of a sha256 as part of the storage key is utilized to help ensure
        uniqueness of bits. The sha256 is also stored with the formula state.

        This relation betewen the formula state and the binary bits, helps
        guarantee the property that any published formula in zookeeper will use
        the binary bits that it was published with.
        """

        yield self.publisher.add_formula("local", self.formula)

        concurrent_publisher = FormulaPublisher(
            self.client, self.storage)

        formula = FormulaDirectory(self.sample_dir1)
        yield concurrent_publisher.add_formula("local", formula)

        yield self.publisher.publish()

        # modify the formula to create a conflict scenario
        self.makeFile("zebra",
                      path = os.path.join(self.sample_dir1, "junk.txt"))

        # assert the formula now has a different sha post modification
        modified_formula_sha = formula.get_sha256()
        self.assertNotEqual(
            modified_formula_sha,
            self.formula.get_sha256())

        # verify publishing raises a stateerror
        def verify_failure(result):
            if not isinstance(result, Failure):
                self.fail("Should have raised state error")
            result.trap(StateChanged)
            return True

        yield concurrent_publisher.publish().addBoth(verify_failure)

        # verify the zk state
        formula_nodes = yield self.client.get_children("/formulas")
        self.assertEqual(formula_nodes, [self.formula_key])

        content, stat = yield self.client.get(
            "/formulas/%s" % self.formula_key)

        # assert the checksum matches the initially published checksum
        self.assertEqual(
            yaml.load(content)["sha256"],
            self.formula_storage_key.split(":")[-1])

        stored_files = os.listdir(os.path.join(self.storage.path, "formulas"))
        self.assertIn(self.formula_storage_key, stored_files)

        # and the binary bits where stored
        modified_formula_storage_key = "local:%s-%s:%s" %(
            self.formula.metadata.name, self.formula.metadata.revision,
            modified_formula_sha)
        self.assertIn(modified_formula_storage_key, stored_files)

    @inlineCallbacks
    def test_add_formula_with_concurrent_removal(self):
        """
        If a formula is published, and it detects that the formula exists
        already exists, it will attempt to retrieve the formula state to
        verify there is no checksum mismatch. If concurrently the formula
        is removed, the publisher should fail with a statechange error.
        """
        formula_id = self.formula.metadata.get_formula_id("local")
        manager = self.mocker.patch(FormulaStateManager)

        manager.get_formula_state(formula_id)
        self.mocker.passthrough()

        def match_formula_bundle(bundle):
            return isinstance(bundle, FormulaBundle)

        manager.add_formula_state("local", MATCH(match_formula_bundle))
        self.mocker.result(fail(zookeeper.NodeExistsException()))

        manager.get_formula_state(formula_id)
        self.mocker.result(fail(zookeeper.NoNodeException()))
        self.mocker.replay()

        yield self.publisher.add_formula("local", self.formula)
        yield self.failUnlessFailure(self.publisher.publish(), StateChanged)

    @inlineCallbacks
    def test_add_formula_already_known(self):
        """Adding an existing formula, is an effective noop, as its not added
        to the internal publisher queue.
        """
        # Do an initial publishing of the formula
        scheduled = yield self.publisher.add_formula("local", self.formula)
        self.assertTrue(scheduled)
        result = yield self.publisher.publish()
        self.assertTrue(result[0].name, self.formula.metadata.name)

        publisher = FormulaPublisher(self.client, self.storage)
        scheduled = yield publisher.add_formula("local", self.formula)
        self.assertFalse(scheduled)
        scheduled = yield publisher.add_formula("local", self.formula)
        self.assertFalse(scheduled)

        result = yield publisher.publish()
        self.assertTrue(result[0].name, self.formula.metadata.name)
        self.assertTrue(result[1].name, self.formula.metadata.name)


class EnvironmentPublisherTest(EnvironmentsConfigTestBase):

    def setUp(self):
        super(EnvironmentPublisherTest, self).setUp()
        self.write_config(SAMPLE_ENV)
        self.config.load()
        self.environment = self.config.get("myfirstenv")
        zookeeper.set_debug_level(0)

    @inlineCallbacks
    def test_publisher_for_environment(self):
        publisher = yield FormulaPublisher.for_environment(self.environment)
        self.assertTrue(isinstance(publisher, FormulaPublisher))
