import os
import errno
import signal
import shutil
import sys

import ensemble

from twisted.internet.defer import Deferred
from twisted.internet.protocol import ProcessProtocol

from ensemble.formula.bundle import FormulaBundle
from ensemble.lib.twistutils import get_module_directory

from .errors import UnitDeploymentError


class AgentProcessProtocol(ProcessProtocol):

    def __init__(self, deferred):
        self.deferred = deferred
        self._error_buffer = []

    def errReceived(self, data):
        self._error_buffer.append(data)

    def processEnded(self, reason):
        if self._error_buffer:
            msg = "".join(self._error_buffer)
            msg.strip()
            self.deferred.errback(UnitDeploymentError(msg))
        else:
            self.deferred.callback(True)


class ServiceUnitDeployment(object):

    unit_agent_module = "ensemble.agents.unit"

    def __init__(self, unit_name, units_directory):
        self.unit_name = unit_name

        assert ".." not in unit_name, "Invalid Unit Name"
        self.unit_path_name = unit_name.replace("/", "-")
        self.units_directory = units_directory

        self.directory = os.path.join(
            self.units_directory, self.unit_path_name)

        self.pid_file = os.path.join(
            self.units_directory, "%s.pid" % self.unit_path_name)

    def start(self, machine_id, zookeeper_hosts):
        """Start a service unit agent."""
        if not os.path.exists(self.directory):
            raise UnitDeploymentError("Formula must be unpacked first.")

        # Start the service unit agent
        log_file = os.path.join(self.directory, "formula.log")

        args = [sys.executable, "-m", self.unit_agent_module, "-n",
                "--pidfile", self.pid_file, "--logfile", log_file]

        environ = dict(os.environ)
        environ["ENSEMBLE_MACHINE_ID"] = str(machine_id)
        environ["ENSEMBLE_UNIT_NAME"] = self.unit_name
        environ["ENSEMBLE_HOME"] = os.path.dirname(self.units_directory)
        environ["ENSEMBLE_ZOOKEEPER"] = zookeeper_hosts
        environ["PYTHONPATH"] = ":".join(
            filter(None, [
                os.path.dirname(get_module_directory(ensemble)),
                environ.get("PYTHONPATH")]))

        from twisted.internet import reactor
        process_deferred = Deferred()
        protocol = AgentProcessProtocol(process_deferred)
        reactor.spawnProcess(protocol, sys.executable, args, environ)
        return process_deferred

    def destroy(self):
        """Forcibly terminate a service unit agent, and clean disk state.

        This will destroy/unmount any state on disk.
        """
        if self.is_running():
            pid = int(open(self.pid_file).read())
            try:
                os.kill(pid, signal.SIGKILL)
            except OSError, e:
                if e.errno != errno.ESRCH:
                    raise

        if os.path.exists(self.pid_file):
            os.remove(self.pid_file)

        if os.path.exists(self.directory):
            shutil.rmtree(self.directory)

    def is_running(self):
        """Is the service unit running."""
        if not os.path.exists(self.pid_file):
            return False

        try:
            pid = int(open(self.pid_file).read())
        except ValueError:
            return False

        # Attempt to send a signal to the process to verify its a valid process
        # From man 2 kill
        # "If sig is 0, then no signal is sent, but error checking is still
        # performed; this can be used to check for the existence of a process
        # ID or process group ID."
        try:
            os.kill(pid, 0)
        except OSError, e:
            if e.errno == errno.ESRCH:
                return False
            raise
        return True

    def unpack_formula(self, formula):
        """Unpack a formula to the service units directory."""
        if not isinstance(formula, FormulaBundle):
            raise UnitDeploymentError(
                "Invalid formula for deployment: %r" % formula)

        formula.extract_to(os.path.join(self.directory, "formula"))
