import copy

from twisted.internet.defer import succeed

from ensemble.errors import EnvironmentNotFound
from ensemble.state.auth import make_identity

from .utils import log

BOOTSTRAP_PACKAGES = [
    "bzr",
    "default-jre-headless",
    "zookeeper",
    "zookeeperd"]


def _get_initialize_script(instance_id_cmd, admin_secret):
    admin_identity = make_identity("admin:%s" % admin_secret)
    script = 'ensemble-admin initialize --instance-id=%s --admin-identity="%s"'
    return script % (instance_id_cmd, admin_identity)

def _get_provision_agent_script(vars):
    script = (
        "ENSEMBLE_ZOOKEEPER=%(ensemble-zookeeper-hosts)s "
        "python -m ensemble.agents.provision -n "
        "--logfile=/var/log/ensemble/provision-agent.log "
        "--pidfile=/var/run/ensemble/provision-agent.pid")
    return script % vars


class Bootstrap(object):
    """Base bootstrap operation class.
    
    You'll need to override C{launch_machine_cls} and
    C{get_instance_id_cmd} to make it useful.
    """

    launch_machine_cls = None

    def __init__(self, provider):
        self._provider = provider

    def run(self):
        """Get an existing zookeeper, or launch a new one.
        
        @return: single-element list containing an appropriate 
            ProviderMachine instance for the zookeeper.
        """
        machines_deferred = self._provider.get_zookeeper_machines()
        machines_deferred.addCallback(self._on_machines_found)
        machines_deferred.addErrback(self._on_no_machines_found)
        return machines_deferred

    def _on_machines_found(self, machines):
        log.info("Ensemble environment previously bootstrapped.")
        return machines

    def _on_no_machines_found(self, failure):
        failure.trap(EnvironmentNotFound)
        log.debug("Launching Ensemble bootstrap instance.")

        machine_data = {"machine-id": "0", "ensemble-zookeeper-hosts": "localhost:2181"}
        bootstrap_launch = self.launch_machine_cls(self._provider)
        return bootstrap_launch.run(
            machine_data,
            pre_start_machine=self._get_bootstrap_variables,
            post_start_machine=self._on_bootstrap_launched)

    def get_instance_id_cmd(self):
        """Snippet to discover local instance-id

        @return: a snippet of shell script that evaluates to the
            local machine's instance_id.
        """
        raise NotImplementedError()

    def _get_bootstrap_variables(self, vars_in):
        vars_out = copy.deepcopy(vars_in)
        vars_out["packages"].extend(BOOTSTRAP_PACKAGES)

        admin_secret = self._provider.config["admin-secret"]
        initialize_script = _get_initialize_script(
            self.get_instance_id_cmd(), admin_secret)
        provision_agent_script = _get_provision_agent_script(vars_out)

        # we assume that the machine agent script is last,
        # and weave other necessary scripts around it :-/.
        # TODO: improve (make vars more structured?)
        scripts = vars_out["scripts"]
        machine_agent_script = scripts.pop()
        scripts.extend([
            initialize_script,
            machine_agent_script,
            provision_agent_script])

        return succeed(vars_out)

    def _on_bootstrap_launched(self, machines):
        instance_ids = [m.instance_id for m in machines]
        d = self._provider.save_state({"zookeeper-instances": instance_ids})
        d.addCallback(lambda _: machines)
        return d

