import copy

from twisted.internet.defer import inlineCallbacks, returnValue, fail

from ensemble.errors import ProviderError

from .utils import get_user_authorized_keys

DEFAULT_REPOSITORIES = [
    "ppa:ensemble/ppa"]

DEFAULT_PACKAGES = [
    "bzr",
    "tmux",
    "python-setuptools",
    "python-twisted",
    "python-argparse",
    "python-txaws",
    "python-zookeeper"]


def _get_branch_install_scripts(branch):
    """Returns the commands to install ensemble from a bzr branch.

    The branch identifier must be a valid publicly accessible bzr
    url that can be used by bzr to checkout the branch. The install
    commands will setup local directories and installation of the branch.
    """
    ensemble_vcs = ("cd /usr/lib/ensemble "
                    "&& sudo /usr/bin/bzr co %s ensemble" % branch)
    ensemble_install = [
        "sudo apt-get install -y python-txzookeeper",
        "sudo mkdir -p /usr/lib/ensemble",
        ensemble_vcs,
        "cd /usr/lib/ensemble/ensemble && sudo python setup.py develop"]

    return ensemble_install

def _get_install_scripts(machine_data):
    if machine_data.get("ensemble-branch"):
        return _get_branch_install_scripts(machine_data["ensemble-branch"])
    return ["sudo apt-get -y install ensemble"]

def _get_launch_scripts(machine_data):
    """Get the scripts that will execute after the machine has launched."""
    launch_scripts = _get_install_scripts(machine_data)

    # These should be in the deb
    launch_scripts.extend([
        "sudo mkdir -p /var/lib/ensemble",
        "sudo mkdir -p /var/log/ensemble"])

    # Every machine has its own agent.
    machine_agent_script = (
        "ENSEMBLE_MACHINE_ID=%(machine-id)s "
        "ENSEMBLE_ZOOKEEPER=%(ensemble-zookeeper-hosts)s "
        "python -m ensemble.agents.machine -n "
        "--logfile=/var/log/ensemble/machine-agent.log "
        "--pidfile=/var/run/ensemble/machine-agent.pid")
    launch_scripts.append(machine_agent_script % machine_data)
    return launch_scripts


class LaunchMachine(object):
    """Abstract class with generic instance-launching logic.

    To create your own subclass, you will certainly need to 
    override C{start_machine}, and may find it convenient
    to override C{get_machine_variables} as well.
    """

    def __init__(self, provider):
        self._provider = provider

    def run(self, machine_data, 
            pre_start_machine=None, 
            post_start_machine=None):
        """Launch an instance node within the machine provider environment.

        @param machine_data a dictionary of data that is passed along to
               provided to the machine as serialized yaml. 'machine-id' is a
               required key, denoting the machine's zookeeper node for its
               machine agent.
        """
        if "machine-id" not in machine_data:
            return fail(ProviderError(
                "Machine state `machine-id` not provided in machine_data."))

        launch_chain = self.get_machine_variables(machine_data)
        if pre_start_machine is not None:
            launch_chain.addCallback(pre_start_machine)
        launch_chain.addCallback(self.start_machine)
        if post_start_machine is not None:
            launch_chain.addCallback(post_start_machine)
        return launch_chain

    @inlineCallbacks
    def get_machine_variables(self, vars_in):
        """Get the variables used for launching a machine and its cloud-init"""
        # If we're using a custom branch, pass it via cloud-init to the agents,
        # so the entire cluster uses it.
        vars_out = copy.deepcopy(vars_in)
        ensemble_branch = self._provider.config.get("ensemble-branch")
        if ensemble_branch:
            vars_out["ensemble-branch"] = ensemble_branch

        # Get the ensemble zookeeper hosts for the node's agents to connect to.
        if "ensemble-zookeeper-hosts" not in vars_out:
            zookeeper_machines = yield self._provider.get_zookeeper_machines()
            zookeeper_hosts = [m.private_dns_name for m in zookeeper_machines]
            # format multiple hosts in string suitable for client
            hosts = ", ".join(["%s:2181" % h for h in zookeeper_hosts])
            vars_out["ensemble-zookeeper-hosts"] = hosts

        authorized_keys = get_user_authorized_keys(self._provider.config)
        vars_out["authorized_keys"] = authorized_keys
        vars_out["packages"] = list(DEFAULT_PACKAGES)
        vars_out["repositories"] = list(DEFAULT_REPOSITORIES)
        vars_out["scripts"] = _get_launch_scripts(vars_out)
        returnValue(vars_out)

    def start_machine(self, variables):
        """Actually launch a machine for the appropriate provider.

        @param variables: non-provider-specific data, sufficient to
            define the machine's behaviour once it exists.

        @return: a list of newly-launched ProviderMachines.
        """
        raise NotImplementedError()


