from twisted.internet.defer import fail, succeed

from ensemble.errors import EnvironmentNotFound, ProviderError
from ensemble.lib.testing import TestCase
from ensemble.providers.common.bootstrap import Bootstrap
from ensemble.providers.common.launch import (
    DEFAULT_PACKAGES, DEFAULT_REPOSITORIES, LaunchMachine)
from ensemble.providers.dummy import DummyMachine

DEFAULT_INSTALL = ["sudo apt-get -y install ensemble"]
CREATE_DIRS = ["sudo mkdir -p /var/lib/ensemble",
               "sudo mkdir -p /var/log/ensemble"]


class DummyLaunchMachine(LaunchMachine):

    def __init__(self, provider):
        self._provider = provider

    def start_machine(self, variables):
        return [DummyMachine("some-instance-id")]


def get_provider(launch_cls, config=None, zookeeper=True):
    class DummyProvider(object):
        config = {"authorized-keys": "abc"}
        
        def get_zookeeper_machines(self):
            if zookeeper:
                return succeed(
                   [DummyMachine("another-instance-id",
                                 private_dns_name="zookeeper.internal")])
            else:
                return fail(EnvironmentNotFound())

    if config is not None:
        DummyProvider.config.update(config)
    return DummyProvider()

def get_launch(cls=DummyLaunchMachine, config=None, zookeeper=True):
    provider = get_provider(cls, config, zookeeper)
    return cls(provider)


class LaunchMachineTest(TestCase):

    def _verify_machines(self, machines):
        (machine,) = machines
        self.assertTrue(isinstance(machine, DummyMachine))
        self.assertEquals(machine.instance_id, "some-instance-id")

    def test_no_machine_id(self):
        launch = get_launch()
        d = launch.run({})
        self.assertFailure(d, ProviderError)
        return d

    def test_basic_run(self):
        launch = get_launch()
        d = launch.run({"machine-id": "machine-77"})
        d.addCallback(self._verify_machines)
        return d

    def test_pre_launch_machine(self):
        original_vars = object()
        tweaked_vars = object()
        def verify_vars(vars):
            self.assertEquals(vars, original_vars)
            return tweaked_vars

        test = self
        original_machine_data = {"machine-id": "machine-32767"}
        class Operation(DummyLaunchMachine):
            def start_machine(self, vars):
                test.assertEquals(vars, tweaked_vars)
                return [DummyMachine("some-instance-id")]
            def get_machine_variables(self, machine_data):
                test.assertEquals(machine_data, original_machine_data)
                return succeed(original_vars)

        launch = get_launch(Operation)
        d = launch.run(original_machine_data,
                       pre_start_machine=verify_vars)
        d.addCallback(self._verify_machines)
        return d

    def test_post_start_machine(self):
        final_result = object()
        def verify_machines(machines):
            self._verify_machines(machines)
            return final_result
        def verify_final_result(result):
            self.assertEquals(result, final_result)
        
        launch = get_launch()
        d = launch.run({"machine-id": "machine-123"},
                       post_start_machine=verify_machines)
        d.addCallback(verify_final_result)
        return d

    def test_get_machine_variables_normal(self):
        launch = get_launch()
        d = launch.get_machine_variables({"machine-id": "machine-999"})
        def verify_vars(vars):
            expect_vars = {
                "authorized_keys": "abc",
                "ensemble-zookeeper-hosts": "zookeeper.internal:2181",
                "machine-id": "machine-999",
                "packages": DEFAULT_PACKAGES,
                "repositories": DEFAULT_REPOSITORIES,
                "scripts": DEFAULT_INSTALL + CREATE_DIRS +
                           ["ENSEMBLE_MACHINE_ID=machine-999 "
                            "ENSEMBLE_ZOOKEEPER=zookeeper.internal:2181 "
                            "python -m ensemble.agents.machine -n "
                            "--logfile=/var/log/ensemble/machine-agent.log "
                            "--pidfile=/var/run/ensemble/machine-agent.pid"]}
            self.assertEquals(vars, expect_vars)
        d.addCallback(verify_vars)
        return d
    
    def test_get_machine_variables_with_preset_zookeeper(self):
        launch = get_launch()
        zookeeper_host =  "another-zookeeper:2181"
        d = launch.get_machine_variables({
            "machine-id": "machine-999",
            "ensemble-zookeeper-hosts": zookeeper_host})
        def verify_vars(vars):
            hosts = vars["ensemble-zookeeper-hosts"]
            self.assertEquals(hosts, zookeeper_host)
            machine_agent_script = vars["scripts"][-1]
            set_zookeeper_host = "ENSEMBLE_ZOOKEEPER=%s" % zookeeper_host
            self.assertTrue(set_zookeeper_host in machine_agent_script)
        d.addCallback(verify_vars)
        return d
    
    def test_get_machine_variables_with_ensemble_branch(self):
        branch_name =  "lp:~my/ensemble/branch"
        launch = get_launch(config={"ensemble-branch": branch_name})
        d = launch.get_machine_variables({"machine-id": "machine-303"})
        def verify_vars(vars):
            self.assertEquals(vars["ensemble-branch"], branch_name)
            install_scripts = ["sudo apt-get install -y python-txzookeeper",
                               "sudo mkdir -p /usr/lib/ensemble",
                               "cd /usr/lib/ensemble && "
                               "sudo /usr/bin/bzr co "
                               + branch_name + " ensemble",
                               "cd /usr/lib/ensemble/ensemble && "
                               "sudo python setup.py develop"]
            self.assertEquals(vars["scripts"][:len(install_scripts)],
                              install_scripts)
        d.addCallback(verify_vars)
        return d


class BootstrapTest(TestCase):

    # TODO move this class into its own file, factor out common dependencies

    def test_bootstrap_variables(self):
        test = self
        class DummyLaunch(DummyLaunchMachine):
            def start_machine(self, vars):
                scripts = vars["scripts"]
                test.assertEquals(scripts[:3], DEFAULT_INSTALL + CREATE_DIRS)
                admin_init, machine_agent, provision_agent = scripts[3:]
                
                test.assertTrue(admin_init.startswith(
                    "ensemble-admin initialize "
                    '--instance-id=$(magic) --admin-identity="admin:'))
                test.assertFalse("SEEKRIT" in admin_init)

                expect_machine = (
                    "ENSEMBLE_MACHINE_ID=0 "
                    "ENSEMBLE_ZOOKEEPER=localhost:2181 "
                    "python -m ensemble.agents.machine -n "
                    "--logfile=/var/log/ensemble/machine-agent.log "
                    "--pidfile=/var/run/ensemble/machine-agent.pid")
                test.assertEquals(machine_agent, expect_machine)

                expect_provision = (
                    "ENSEMBLE_ZOOKEEPER=localhost:2181 "
                    "python -m ensemble.agents.provision -n "
                    "--logfile=/var/log/ensemble/provision-agent.log "
                    "--pidfile=/var/run/ensemble/provision-agent.pid")
                test.assertEquals(provision_agent, expect_provision)

                return [DummyMachine("an-instance-id")]
        provider = get_provider(
            DummyLaunch, config={"admin-secret": "SEEKRIT"}, zookeeper=False)
        saved_state = {}
        provider.save_state = lambda s: succeed(saved_state.update(s))

        class DummyBootstrap(Bootstrap):
            launch_machine_cls = DummyLaunch
            def get_instance_id_cmd(self):
                return "$(magic)"
        bootstrap = DummyBootstrap(provider)

        d = bootstrap.run()
        def verify_machines(machines):
            (machine,) = machines
            self.assertTrue(isinstance(machine, DummyMachine))
            self.assertEquals(machine.instance_id, "an-instance-id")
            self.assertEquals(saved_state["zookeeper-instances"], ["an-instance-id"])
        d.addCallback(verify_machines)
        return d
