from yaml import dump

from twisted.internet.defer import succeed

from txaws.s3.client import S3Client
from txaws.ec2.client import EC2Client
from txaws.ec2.model import Instance, SecurityGroup

from ensemble.lib.mocker import KWARGS, MATCH
from ensemble.providers.ec2 import MachineProvider


MATCH_AMI = MATCH(lambda image_id: image_id.startswith("ami-"))


class EC2TestMixin(object):

    env_name = "moon"
    service_factory_kwargs = None

    def get_config(self):
        return {"type": "ec2",
                "admin-secret": "magic-beans",
                "control-bucket": self.env_name}

    def get_provider(self):
        """Return the ec2 machine provider.

        This should only be invoked after mocker is in replay mode so the
        AWS service class will be appropriately replaced by the mock.
        """
        return MachineProvider(self.env_name, self.get_config())

    def setUp(self):
        # mock out the aws services
        service_factory = self.mocker.replace(
            "txaws.service.AWSServiceRegion")
        self._service = service_factory(KWARGS)

        def store_factory_kwargs(**kwargs):
            self.service_factory_kwargs = kwargs
        self.mocker.call(store_factory_kwargs)
        self.s3 = self.mocker.mock(S3Client)
        self._service.get_s3_client()
        self.mocker.result(self.s3)
        self.ec2 = self.mocker.mock(EC2Client)
        self._service.get_ec2_client()
        self.mocker.result(self.ec2)


class EC2MachineLaunchMixin(object):

    default_zookeeper_host = Instance(
        "i-es-zoo", "running", private_dns_name="es.example.internal")

    def _mock_launch_utils(self):
        get_public_key = self.mocker.replace(
            "ensemble.providers.common.utils.get_user_authorized_keys")

        def match_config(arg):
            return isinstance(arg, dict)

        get_public_key(MATCH(match_config))
        self.mocker.result("zebra")
        # AMI is not retrieve dynamically at the moment, default values are
        # utilized.
        #get_ami = self.mocker.replace(
        #    "ensemble.providers.ec2.utils.get_current_ami")
        #get_ami()
        #self.mocker.result(succeed("ami-714ba518"))

    def _mock_create_group(self):
        group_name = "ensemble-%s" % self.env_name
        self.ec2.create_security_group(
            group_name, "Ensemble group for %s" % self.env_name)
        self.mocker.result(succeed(True))

        self.ec2.authorize_security_group(
            group_name, ip_protocol="tcp", from_port="22",
            to_port="22", cidr_ip="0.0.0.0/0")
        self.mocker.result(succeed([self.env_name]))

        self.ec2.describe_security_groups(group_name)
        self.mocker.result(succeed(
            [SecurityGroup(group_name, "", owner_id="123")]))

        self.ec2.authorize_security_group(
            group_name, source_group_name=group_name,
            source_group_owner_id="123")
        self.mocker.result(succeed(True))

        self.ec2.authorize_security_group(
            group_name,
            ip_protocol="tcp",
            from_port="1",
            to_port="2180",
            cidr_ip="0.0.0.0/0")
        self.ec2.authorize_security_group(
            group_name,
            ip_protocol="tcp",
            from_port="2182",
            to_port="65535",
            cidr_ip="0.0.0.0/0")

        self.ec2.authorize_security_group(
            group_name,
            ip_protocol="udp",
            from_port="1",
            to_port="2180",
            cidr_ip="0.0.0.0/0")
        self.ec2.authorize_security_group(
            group_name,
            ip_protocol="udp",
            from_port="2182",
            to_port="65535",
            cidr_ip="0.0.0.0/0")

        self.mocker.result(succeed(True))

    def _mock_get_zookeeper_hosts(self, hosts=None):
        """
        Try to encapsulate a variety of behaviors here..

        if hosts is None, a default host is used.
        if hosts is False, no s3 state is returned
        if hosts are passed as a list of instances, they
           are returned.
        """

        if hosts is None:
            hosts = [self.default_zookeeper_host]

        self.s3.get_object(self.env_name, "provider-state")
        if hosts is False:
            self.mocker.result(succeed(None))
            return

        state = dump({
            "zookeeper-instances":
            [i.instance_id for i in hosts]})

        self.mocker.result(succeed(state))
        if hosts:
            # connect grabs the first host of a set.
            self.ec2.describe_instances(hosts[0].instance_id)
            self.mocker.result(succeed([hosts[0]]))


