import logging

from yaml import dump, load

from twisted.internet.defer import succeed

from txaws.ec2.model import Instance, SecurityGroup

from ensemble.errors import EnvironmentPending
from ensemble.lib.mocker import MATCH
from ensemble.lib.testing import TestCase
from ensemble.providers.common.bootstrap import BOOTSTRAP_PACKAGES
from ensemble.providers.common.launch import (
    DEFAULT_REPOSITORIES, DEFAULT_PACKAGES)
from ensemble.providers.ec2.machine import EC2ProviderMachine
from ensemble.state.auth import make_identity

from .common import EC2TestMixin, EC2MachineLaunchMixin, MATCH_AMI


class EC2BootstrapTest(EC2TestMixin, EC2MachineLaunchMixin, TestCase):

    def _mock_save(self):
        """Mock saving bootstrap instances to S3."""
        self.s3.get_bucket(self.env_name)
        self.mocker.result(succeed(True))

        def match_string(data):
            return isinstance(data, str)

        self.s3.put_object(
            self.env_name, "provider-state",
            MATCH(match_string),
            content_type="text/plain")
        self.mocker.result(succeed(True))

    def _mock_launch(self):
        """Mock launching a bootstrap machine on ec2."""
        credentials = "admin:%s" % self.get_config()["admin-secret"]
        admin_identity = make_identity(credentials)

        def verify_user_data(data):

            lines = data.split("\n")
            self.assertEqual(lines.pop(0), "#cloud-config")
            config = load("\n".join(lines))
            repos = [dict(source=r) for r in DEFAULT_REPOSITORIES]
            self.assertEqual(config["apt_sources"], repos)
            self.assertEqual(
                config["packages"],
                list(DEFAULT_PACKAGES) + BOOTSTRAP_PACKAGES)
            self.failUnlessIn("admin-identity", config["runcmd"][-3])

            script = (
                'ensemble-admin initialize --instance-id=%s'
                ' --admin-identity="%s"') % (
                    "$(curl http://169.254.169.254/1.0/meta-data/instance-id)",
                    admin_identity)

            self.assertEqual(config["runcmd"][-3], script)

            script = (
                "ENSEMBLE_MACHINE_ID=0 ENSEMBLE_ZOOKEEPER=localhost:2181 "
                "python -m ensemble.agents.machine -n "
                "--logfile=/var/log/ensemble/machine-agent.log "
                "--pidfile=/var/run/ensemble/machine-agent.pid")

            self.assertEqual(config["runcmd"][-2], script)

            provision_agent_script = (
                "ENSEMBLE_ZOOKEEPER=localhost:2181 "
                "python -m ensemble.agents.provision -n "
                "--logfile=/var/log/ensemble/provision-agent.log "
                "--pidfile=/var/run/ensemble/provision-agent.pid")
            self.assertEqual(config["runcmd"][-1], provision_agent_script)
            return True

        self.ec2.run_instances(
            image_id=MATCH_AMI,
            instance_type="m1.small",
            max_count=1,
            min_count=1,
            security_groups=["ensemble-%s" % self.env_name],
            user_data=MATCH(verify_user_data))

    def test_launch_bootstrap(self):
        """The provider bootstrap can launch a bootstrap/zookeeper machine."""

        log = self.capture_logging("ensemble.common", level=logging.DEBUG)

        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(None))
        self.ec2.describe_security_groups()
        self.mocker.result(succeed([]))
        self._mock_create_group()
        self._mock_launch_utils()
        self._mock_launch()
        self.mocker.result(succeed([]))
        self._mock_save()
        self.mocker.replay()

        provider = self.get_provider()
        deferred = provider.bootstrap()

        def check_log(result):
            log_text = log.getvalue()
            self.assertIn("Launching Ensemble bootstrap instance", log_text)
            self.assertNotIn("previously bootstrapped", log_text)

        deferred.addCallback(check_log)
        return deferred

    def test_launch_bootstrap_existing_provider_group(self):
        """
        When launching a bootstrap instance the provider will use an existing
        provider instance group.
        """
        self.capture_logging("ensemble.ec2")
        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(None))
        self.ec2.describe_security_groups()
        self.mocker.result(succeed([
            SecurityGroup("ensemble-%s" % self.env_name, "")]))
        self._mock_launch_utils()
        self._mock_launch()
        self.mocker.result(succeed([]))
        self._mock_save()
        self.mocker.replay()

        provider = self.get_provider()
        return provider.bootstrap()

    def test_run_with_loaded_state(self):
        """
        If the provider bootstrap is run when there is already a running
        bootstrap instance, it will just return the existing machine.
        """
        instances = [
            Instance("i-foobar", "running", dns_name="x1.example.com")]
        state = dump({
            "zookeeper-instances":
            [i.instance_id for i in instances]})
        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(state))
        self.ec2.describe_instances(*[i.instance_id for i in instances])
        self.mocker.result(succeed(instances))
        self.mocker.replay()

        log = self.capture_logging("ensemble.common")

        def validate_result(result):
            self.assertTrue(result)
            machine = result.pop()
            self.assertTrue(isinstance(machine, EC2ProviderMachine))
            self.assertEqual(machine.instance_id,
                             instances[0].instance_id)
            self.assertEquals(
                log.getvalue(),
                "Ensemble environment previously bootstrapped.\n")

        provider = self.get_provider()
        d = provider.bootstrap()
        d.addCallback(validate_result)
        return d

    def test_run_with_pending(self):
        """
        If the environment is not running, but is in a pending state,
        the underlying error from the connection attempt will pop up.
        """
        instances = [
            Instance("i-foobar", "pending", dns_name="x1.example.com")]
        state = dump({
            "zookeeper-instances":
            [i.instance_id for i in instances]})
        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(state))
        self.ec2.describe_instances(*[i.instance_id for i in instances])
        self.mocker.result(succeed(instances))
        self.mocker.replay()

        provider = self.get_provider()
        d = provider.bootstrap()
        self.assertFailure(d, EnvironmentPending)
        return d

    def test_run_with_launch(self):
        """
        The provider bootstrap will launch an instance when run if there
        is no existing instance.
        """
        instances = [
            Instance("i-foobar", "running", dns_name="x1.example.com")]

        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(None))
        self.ec2.describe_security_groups()
        self.mocker.result(succeed([
            SecurityGroup("ensemble-%s" % self.env_name, "")]))
        self._mock_launch_utils()
        self._mock_launch()
        self.mocker.result(succeed(instances))
        self._mock_save()
        self.mocker.replay()

        def validate_result(result):
            self.assertEqual(len(result), 1)
            result = result.pop()
            self.assertTrue(isinstance(result, EC2ProviderMachine))
            self.assertEqual(result.instance_id,
                             instances[0].instance_id)

        provider = self.get_provider()
        d = provider.bootstrap()
        d.addCallback(validate_result)
        return d

    def test_provider_bootstrap(self):
        """The ec2 provider facade instantiates and runs the bootstrap."""
        instances = [
            Instance("i-foobar", "running", dns_name="x1.example.com")]
        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(None))
        self.ec2.describe_security_groups()
        self.mocker.result(succeed([
            SecurityGroup("ensemble-%s" % self.env_name, "")]))
        self._mock_launch_utils()
        self._mock_launch()
        self.mocker.result(succeed(instances))
        self._mock_save()
        self.mocker.replay()

        def validate_result(result):
            self.assertEqual(len(result), 1)
            result = result.pop()
            self.assertTrue(isinstance(result, EC2ProviderMachine))
            self.assertEqual(result.instance_id,
                             instances[0].instance_id)

        provider = self.get_provider()
        d = provider.bootstrap()
        d.addCallback(validate_result)
        return d

