from twisted.internet.defer import fail, succeed

from txaws.ec2.model import Instance
from txaws.ec2.exception import EC2Error

from yaml import dump

from ensemble.errors import EnvironmentNotFound, EnvironmentPending
from ensemble.lib.testing import TestCase
from ensemble.providers.ec2.machine import EC2ProviderMachine
from ensemble.providers.ec2.tests.common import EC2TestMixin


def _invalid_id_error():
    e = EC2Error("<error/>", 400)
    e.errors = [{"Code": "InvalidInstanceID.NotFound"}]
    return e


class EC2FindZookeepersTest(EC2TestMixin, TestCase):

    def test_get_zookeeper_machines_no_state(self):
        """
        When loading saved state from S3, the provider method gracefully
        handles the scenario where there is no saved state.
        """
        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(dump([])))
        self.mocker.replay()

        provider = self.get_provider()
        d = provider.get_zookeeper_machines()
        self.failUnlessFailure(d, EnvironmentNotFound)
        return d

    def test_get_zookeeper_machines_state_no_hosts(self):
        """
        If the saved state from s3 exists, but has no zookeeper hosts,
        the provider method correctly detects this and raises
        EnvironmentNotFound.
        """
        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(dump({"abc": 123})))
        self.mocker.replay()

        provider = self.get_provider()
        d = provider.get_zookeeper_machines()
        self.failUnlessFailure(d, EnvironmentNotFound)
        return d

    def test_get_zookeeper_machines_not_running(self):
        """
        If the saved state exists but only contains zookeeper hosts that
        are not actually running, the provider method detects this and raises 
        EnvironmentNotFound.
        """
        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(dump(
            {"zookeeper-instances": ["i-x"]})))
        self.ec2.describe_instances(*["i-x"])
        self.mocker.result(succeed([]))
        self.mocker.replay()

        provider = self.get_provider()
        d = provider.get_zookeeper_machines()
        self.failUnlessFailure(d, EnvironmentNotFound)
        return d

    def test_get_zookeeper_machines_with_invalid_instance_id(self):
        """
        If the saved state only contins zookeeper hosts with invalid 
        instance ids, the provider method detects this and raises
        EnvironmentNotFound
        """
        server1 = Instance("i-foobar", "pending", dns_name="x1.example.com")
        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(dump(
            {"zookeeper-instances": [server1.instance_id]})))

        self.ec2.describe_instances(server1.instance_id)
        self.mocker.result(fail(_invalid_id_error()))
        self.mocker.replay()

        provider = self.get_provider()
        d = provider.get_zookeeper_machines()
        return self.assertFailure(d, EnvironmentNotFound)

    def test_get_zookeeper_machines_with_pending(self):
        """
        If the only zookeeper instances are still pending, the 
        provider method detects this and raises EnvironmentPending.
        """
        server = Instance("i-foobar", "pending", dns_name="x1.example.com")
        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(dump(
            {"zookeeper-instances": [server.instance_id]})))
        self.ec2.describe_instances(*[server.instance_id])
        self.mocker.result(succeed([server]))
        self.mocker.replay()

        provider = self.get_provider()
        d = provider.get_zookeeper_machines()
        self.assertFailure(d, EnvironmentPending)

        def check_error(error):
            self.assertEquals(str(error),
                              "Started machine is still pending.")
        d.addCallback(check_error)
        return d

    def test_get_zookeeper_machines_eventual_success(self):
        """
        When the S3 state contains running zookeeper hosts,
        return a one-element list containing the first one 
        encountered.
        """
        server1 = Instance("i-foobar", "pending", dns_name="x1.example.com")
        server2 = Instance("i-foobaz", "running", dns_name="x2.example.com")
        server3 = Instance("i-foobop", "running", dns_name="x3.example.com")
        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(dump(
            {"zookeeper-instances": ["i-invalid",
                                     server1.instance_id,
                                     server2.instance_id,
                                     server3.instance_id]})))

        # Zk instances are checked individually to handle invalid ids correctly
        self.ec2.describe_instances("i-invalid")
        self.mocker.result(fail(_invalid_id_error()))

        self.ec2.describe_instances(server1.instance_id)
        self.mocker.result(succeed([server1]))

        self.ec2.describe_instances(server2.instance_id)
        self.mocker.result(succeed([server2]))

        self.mocker.replay()

        provider = self.get_provider()
        d = provider.get_zookeeper_machines()

        def verify_machines(machines):
            (machine,) = machines
            self.assertEquals(machine.instance_id, "i-foobaz")
            self.assertTrue(isinstance(machine, EC2ProviderMachine))
        d.addCallback(verify_machines)
        return d

