
from yaml import load

from twisted.internet.defer import succeed

from txaws.ec2.model import Instance

from ensemble.errors import EnvironmentNotFound
from ensemble.providers.common.launch import (
    DEFAULT_REPOSITORIES, DEFAULT_PACKAGES)
from ensemble.providers.ec2.machine import EC2ProviderMachine

from ensemble.lib.testing import TestCase
from ensemble.lib.mocker import MATCH

from .common import EC2TestMixin, EC2MachineLaunchMixin, MATCH_AMI


class EC2MachineLaunchTest(EC2TestMixin, EC2MachineLaunchMixin, TestCase):

    def _mock_launch(self, instance=None, custom_verify=None):

        def verify_user_data(data):
            lines = data.split("\n")
            self.assertEqual(lines.pop(0), "#cloud-config")
            config = load("\n".join(lines))
            repos = [dict(source=r) for r in DEFAULT_REPOSITORIES]
            self.assertEqual(config["apt_sources"], repos)
            self.assertEqual(config["packages"], DEFAULT_PACKAGES)
            self.assertEqual(config["machine-data"]["machine-id"], "machine-1")

            if custom_verify:
                custom_verify(config)
            return True

        self.ec2.run_instances(
            image_id=MATCH_AMI,
            instance_type="m1.small",
            max_count=1,
            min_count=1,
            security_groups=["%s-%s" % ("ensemble", self.env_name)],
            user_data=MATCH(verify_user_data))

        if instance:
            self.mocker.result(succeed([instance]))
        else:
            self.mocker.result(succeed([]))

    def test_provider_launch(self):
        """
        The provider can be used to launch a machine with a minimal set of
        required packages, repositories, and and security groups.
        """
        instance = Instance("i-foobar", "running", dns_name="x1.example.com")

        self.ec2.describe_security_groups()
        self.mocker.result(succeed([]))
        self._mock_create_group()
        self._mock_launch_utils()
        self._mock_get_zookeeper_hosts()
        self._mock_launch(instance)
        self.mocker.replay()

        def verify_result(result):
            self.assertEqual(len(result), 1)
            result = result.pop()
            self.assertTrue(isinstance(result, EC2ProviderMachine))
            self.assertEqual(result.instance_id, instance.instance_id)

        provider = self.get_provider()
        d = provider.start_machine({"machine-id": "machine-1"})
        d.addCallback(verify_result)
        return d

    def test_provider_type_machine_variable(self):
        """The provider type is available via cloud-init."""
        self.ec2.describe_security_groups()
        self.mocker.result(succeed([]))
        self._mock_create_group()
        self._mock_launch_utils()
        self._mock_get_zookeeper_hosts()

        def verify_provider_type(data):
            self.assertEqual(
                data["machine-data"]["ensemble-provider-type"], "ec2")

        self._mock_launch(custom_verify=verify_provider_type)
        self.mocker.replay()

        provider = self.get_provider()
        d = provider.start_machine({"machine-id": "machine-1"})
        return d

    def test_launch_with_no_ensemble_s3_state(self):
        # TODO not really testing anything EC2-specific
        """
        Attempting to launch without any ensemble saved state, means
        we can't provide a way for a launched instance to connect
        to the zookeeper shared state. An assertion error is raised
        instead of allowing this.
        """
        self._mock_get_zookeeper_hosts(False)
        self.mocker.replay()

        provider = self.get_provider()
        d = provider.start_machine({"machine-id": "machine-1"})
        self.assertFailure(d, EnvironmentNotFound)
        return d

    def test_launch_with_no_ensemble_zookeeper_hosts(self):
        # TODO not really testing anything EC2-specific
        """
        Attempting to launch without any ensemble zookeeper hosts, means
        we can't provide a way for a launched instance to connect
        to the zookeeper shared state. An assertion error is raised
        instead of allowing this.
        """
        self._mock_get_zookeeper_hosts([])
        self.mocker.replay()

        provider = self.get_provider()
        d = provider.start_machine({"machine-id": "machine-1"})
        self.assertFailure(d, EnvironmentNotFound)
        return d

    def test_launch_starts_machine_agent(self):
        # TODO not really testing anything EC2-specific
        """A launch'd machine should have a machine agent."""
        self.ec2.describe_security_groups()
        self.mocker.result(succeed([]))
        self._mock_create_group()
        self._mock_launch_utils()
        self._mock_get_zookeeper_hosts()

        def verify_machine_agent(data):
            script = ("ENSEMBLE_MACHINE_ID=machine-1 "
                      "ENSEMBLE_ZOOKEEPER=es.example.internal:2181 "
                      "python -m ensemble.agents.machine "
                      "-n --logfile=/var/log/ensemble/machine-agent.log "
                      "--pidfile=/var/run/ensemble/machine-agent.pid")
            self.assertEqual(script, data["runcmd"][-1])

        self._mock_launch(custom_verify=verify_machine_agent)
        self.mocker.replay()

        provider = self.get_provider()
        d = provider.start_machine({"machine-id": "machine-1"})
        return d

