from twisted.internet.defer import fail

from ensemble.errors import EnvironmentNotFound

from .bootstrap import OrchestraBootstrap

from .cobbler import CobblerControl

from ensemble.lib.httpstorage import HttpStorage
from .state import OrchestraSaveState, OrchestraLoadState
from .launch import OrchestraLaunchMachine

#####################
from txzookeeper import ZookeeperClient
from twisted.internet.defer import succeed, fail

from .machine import OrchestraMachine

from ensemble.errors import (
    FileNotFound, ProviderError, ProviderInteractionError)

class MachineProvider(object):

    def __init__(self, environment_name, config):
        self.environment_name = environment_name
        self.config = config
        self.config['instance-id-expression']='$(cat /etc/instance-id)'
        self.config['storage-url'] = ('http://%s/webdav' % self.config['orchestra-server'])
        #if control is None:
        control = CobblerControl(self.config)
        self.control = control
        self._state = None
        self._storage = None

    @property
    def _zk_hosts(self):
        #return self.config.get('zookeeper-hosts',
        #        '%s:%s' % (self.config['orchestra-server'],
        #            self.config.get('zookeeper-port', '2181')))
        # FIXME: use here address of the zookeeper server, not orchestra
        return self.config.get('zookeeper-hosts',
                '%s:%s' % ("192.168.123.101",
                    self.config.get('zookeeper-port', '2181')))

    # FIXME: To be able to show the machines listed in <ensemble status>
    #  we need to pass to ks_meta machine_id=$system_name
    @property
    def _machines(self):
        """List all the machine running in the provider."""
        # FIXME: This doesn't seem to be needed anymore as orchestra-server is no longer
        # the zookeeper
        #machines = [OrchestraMachine('orchestra-server',
        #        self.config['orchestra-server'])]
        machines = []
        for system in self.control.get_acquired_systems():
            #log.debug("acquired system %s" % system)
            meta = self.control.get_system_ks_meta(system)
            #log.debug("acquired system %s meta = %s" % (system, meta))
            try:
                machine_id = str(meta['machine_id'])
            except KeyError:
                #FIXME: Do we really need do this??
                #  had to add machine_id=XYZ to ks_meta to not raise this Exception
                raise Exception(str(system)+str(meta))
            machines.append(OrchestraMachine(system,system))

        #log.debug("machines = %s" % [x.instance_id for x in machines])
        return machines

    @property
    def _admin_id(self):
        return make_identity('admin:%s' % self.config['admin-secret'])

    def connect(self, share=False):
        """
        Connect to the zookeeper ensemble running in the machine provider.

        @param share: Requests sharing of the connection with other clients
            attempting to connect to the same provider, if that's feasible.

        returns an open C{txzookeeper.client.ZookeeperClient} and a
        C{ensemble.storage.connection.TunnelProtocol}
        """
        print "MachineProvider - Connect"
        return ZookeeperClient(self._zk_hosts,
                session_timeout=1000).connect()
        raise NotImplementedError()

    def list_machines(self):
        """
        List all machines running in the provider.
        """
        print "MachineProvider - list_machines"
        return succeed(self._machines)
        raise NotImplementedError()
        
    def start_machine(self, machine_data):
        """
        Start a machine in the provider.

        @param machine_data a dictionary of data to pass along to the newly
               launched machine.
        @type dict
        """
        print "MachineProvider - start_machine"
        launch = OrchestraLaunchMachine(self)
        return launch.run(machine_data=machine_data)
        raise NotImplementedError()

    def get_machine(self, provider_machine_id):
        """
        Retrieve a provider machine by provider machine id.
        """
        print "MachineProvider - get_machine"
        for machine in self._machines:
            if str(provider_machine_id) == machine.instance_id:
                return succeed(machine)
        return fail(ProviderInteractionError("Machine (id: %s) not found" % (provider_machine_id)))
        raise NotImplementedError()

    def bootstrap(self):
        """
        Bootstrap an ensemble server in the provider.
        """
        bootstrap = OrchestraBootstrap(self)
        return bootstrap.run()

    def shutdown(self):
        """
        Terminate any machine resources associated with the provider.
        """
        print "MachineProvider - shutdown"
        machines = self._machines
        for machine in machines:
            if machine.instance_id != 'orchestra-server':
                self.shutdown_machine(machine)

        return succeed(machines)
        raise NotImplementedError()

    def shutdown_machine(self, machine):
        """
        Stop a machine in the provider.
        """
        print "MachineProvider - shutdown_machine"
        if not isinstance(machine, OrchestraMachine):
            return fail(ProviderError("Invalid machine for provider"))

        if machine.dns_name == 'orchestra-server':
            return succeed(machine)

        for m in self._machines:
            if machine.instance_id == m.instance_id:
                self.control.release_system(machine.instance_id)
                return succeed(machine)
        raise NotImplementedError()

    def save_state(self, state):
        """
        Save state to the provider.

        @param state
        @type dict
        """
        print "MachineProvider - save_state"
        serializer = OrchestraSaveState(self)
        return serializer.run(state)
        raise NotImplementedError()

    def load_state(self):
        """
        Load state from the provider, returns a dictionary.
        """
        print "MachineProvider - load_state"
        deserializer = OrchestraLoadState(self)
        return deserializer.run()
        raise NotImplementedError()

    def get_file_storage(self):
        """Retrieve the provider C{FileStorage} abstraction."""
        print "MachineProvider - get_file_storage"
        if self._storage:
            return self._storage
        storage_path = self.config.get("storage-url")
        if storage_path is None:
            storage_path = tempfile.mkdtemp()
        self._storage = HttpStorage(storage_path)
        return self._storage
        raise NotImplementedError()

    def get_serialization_data(self):
        """Return a dictionary serialization of the provider configuration.

        Additionally this extracts credential information from the environment.
        """
        print "MachineProvider - get_serialization_data"
        config = self.config.copy()
        return config
        raise NotImplementedError()

    def open_port(self, machine, port, protocol="tcp"):
        """Expose port to the environment.

        Approximate equivalent of ec2-authorize-group
        """
        raise NotImplementedError()

    def close_port(self, machine, port, protocol="tcp"):
        """Close opened port

        Approximate equivalent of ec2-revoke-group
        """
        raise NotImplementedError()

    def get_opened_ports(self, machine):
        """List ports currently exposed to the environment.

        Approximate equivalent of ec2-describe-group
        """
        raise NotImplementedError()

    def get_zookeeper_machines(self):
        """Find running zookeeper instances.

        @return: the first valid instance found as a single element list.

        @raise: EnvironmentNotFound or EnvironmentPending
        """
        return fail(EnvironmentNotFound())


class FileStorage(object):

    def get(self, name):
        """Returns open file or fails with FileNotFound"""
        raise NotImplementedError()

    def put(self, remote_path, local_path):
        """Copies contents of local_path to remote_path"""
        raise NotImplementedError()

