import yaml

from twisted.internet.defer import (
    inlineCallbacks, returnValue, succeed)

from zookeeper import NoNodeException

from ensemble.formula.config import ConfigOptions
from ensemble.formula.metadata import MetaData
from ensemble.state.base import StateBase
from ensemble.state.errors import FormulaStateNotFound


class FormulaStateManager(StateBase):
    """Manages the state of formulas in an environment."""

    @inlineCallbacks
    def add_formula_state(self, namespace, formula):
        """Register metadata about the provided Formula.

        @param namespace: The namespace this formula will be
            registered under.
        """
        name = formula.metadata.name
        revision = formula.metadata.revision

        formula_data = {
            "config": formula.config.get_serialization_data(),
            "metadata": formula.metadata.get_serialization_data(),
            "sha256": formula.get_sha256(),
        }

        formula_id = "%s:%s-%s" % (namespace, name, revision)

        # XXX In the future we'll have to think about formula
        #     replacements here. For now this will do, and will
        #     explode reliably in case of conflicts.
        yield self._client.create("/formulas/%s" % formula_id,
                                  yaml.safe_dump(formula_data))

        formula_state = FormulaState(self._client, namespace,
                                     name, revision, formula_data)
        returnValue(formula_state)

    @inlineCallbacks
    def get_formula_state(self, formula_id):
        """Retrieve a FormulaState for the given formula id."""
        try:
            content, stat = yield self._client.get("/formulas/" + formula_id)
        except NoNodeException:
            raise FormulaStateNotFound(formula_id)
        formula_data = yaml.load(content)
        rest, revision = formula_id.rsplit("-", 1)
        revision = int(revision)
        namespace, name = rest.split(":", 1)
        formula_state = FormulaState(self._client, namespace,
                                     name, revision, formula_data)
        returnValue(formula_state)


class FormulaState(object):
    """State of a formula registered in an environment."""

    def __init__(self, client, namespace, name, revision,
                 formula_data):
        self._client = client
        self._namespace = namespace
        self._name = name
        self._revision = revision

        self._metadata = MetaData()
        self._metadata.parse_serialization_data(formula_data["metadata"])

        self._config = ConfigOptions()
        self._config.parse(formula_data["config"])

        # Just a health check:
        assert self._metadata.name == name
        assert self._metadata.revision == revision

        self._sha256 = formula_data["sha256"]

    @property
    def namespace(self):
        """The formula namespace."""
        return self._namespace

    @property
    def name(self):
        """The formula name."""
        return self._name

    @property
    def revision(self):
        """The monotonically increasing formula revision number.
        """
        return self._revision

    @property
    def id(self):
        """The formula identifier, of the form <namespace>:<name>-<revision>.
        """
        return "%s:%s-%s" % \
               (self._namespace, self._name, self._revision)

    def get_metadata(self):
        """Return deferred MetaData."""
        return succeed(self._metadata)

    def get_config(self):
        """Return deferred ConfigOptions."""
        return succeed(self._config)

    def get_sha256(self):
        """Return deferred sha256 for the formula."""
        return succeed(self._sha256)

